"use strict";
/// <reference path="../lrt.d.ts" />
var NodeHTTP = require("http");
var NodeURLParser = require("url");
/**
 * The status types for HTTP server.
 */
(function (ServerStatus) {
    /**
     * Server is not working.
     */
    ServerStatus[ServerStatus["IDLE"] = 0] = "IDLE";
    /**
     * Server is working normally.
     */
    ServerStatus[ServerStatus["WORKING"] = 1] = "WORKING";
    /**
     * Server is waiting to be shutdown.
     */
    ServerStatus[ServerStatus["CLOSING"] = 2] = "CLOSING";
    /**
     * Server has been shutdown.
     */
    ServerStatus[ServerStatus["CLOSED"] = 3] = "CLOSED";
    /**
     * Server was in wrong situation.
     */
    ServerStatus[ServerStatus["ERROR"] = 255] = "ERROR";
})(exports.ServerStatus || (exports.ServerStatus = {}));
var ServerStatus = exports.ServerStatus;
/**
 * Remove the last "/" of path and add first "/" if not exists.
 */
function fixPath(path) {
    if (path[0] !== "/") {
        path = "/" + path;
    }
    if (path[path.length - 1] === "/") {
        return path.substr(0, path.length - 1);
    }
    return path;
}
var Server = (function () {
    function Server(options) {
        this.options = {
            "host": "127.0.0.1",
            "port": 80,
            "backlog": 511
        };
        this.serverStatus = ServerStatus.IDLE;
        this.pathStringMap = {};
        this.pathRegExpMap = [];
        if (options) {
            for (var key in this.options) {
                options[key] && (this.options[key] = options[key]);
            }
        }
    }
    Object.defineProperty(Server.prototype, "status", {
        get: function () {
            return this.serverStatus;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Start the HTTP Server.
     */
    Server.prototype.start = function (cb) {
        var callback = (function (err) {
            if (err) {
                this.serverStatus = ServerStatus.ERROR;
                console.log("HTTP Server " + this.options.host + ":" + this.options.port + ": Server started failed.");
                cb && cb(err);
                return;
            }
            console.log("HTTP Server " + this.options.host + ":" + this.options.port + ": Server started okay.");
            this.serverStatus = ServerStatus.WORKING;
            cb && cb(err);
        }).bind(this);
        if (this.serverStatus !== ServerStatus.IDLE) {
            return cb && cb({
                "name": "BAD-STATUS",
                "message": "Server has already started."
            });
        }
        this.serverInst = NodeHTTP.createServer(this.requestEntry.bind(this));
        this.serverInst.listen(this.options.port, this.options.host, this.options.backlog, callback);
        console.log("HTTP Server " + this.options.host + ":" + this.options.port + ": Server is starting...");
    };
    Server.prototype.close = function () {
        if (this.status === ServerStatus.WORKING) {
            this.serverInst.close((function () {
                this.serverStatus = ServerStatus.CLOSED;
                console.log("HTTP Server " + this.options.host + ":" + this.options.port + ": Server has been shutdown.");
            }).bind(this));
            this.serverStatus = ServerStatus.CLOSING;
        }
        console.log("HTTP Server " + this.options.host + ":" + this.options.port + ": Server is shutting down...");
    };
    Server.prototype.requestEntry = function (request, response) {
        if (this.status === ServerStatus.CLOSING) {
            response.setHeader("Connection", "Close");
            response.statusCode = 503;
            response.statusMessage = "Server is Unavailable";
            response.end();
            request.socket.destroy();
            return;
        }
        var urlInfo = NodeURLParser.parse(request.url);
        var handlerFound = false;
        request.path = urlInfo.pathname ? fixPath(urlInfo.pathname) : "/";
        request.query = {};
        for (var path in this.pathStringMap) {
            if (request.path === path) {
                handlerFound = true;
                setTimeout(this.pathStringMap[path], 0, request, response, this);
                break;
            }
        }
        if (!handlerFound) {
            for (var _i = 0, _a = this.pathRegExpMap; _i < _a.length; _i++) {
                var mapRule = _a[_i];
                if (request.pathMatch = request.path.match(mapRule.expr)) {
                    handlerFound = true;
                    setTimeout(mapRule.handler, 0, request, response, this);
                    break;
                }
            }
        }
        if (!handlerFound) {
            response.statusCode = 404;
            response.statusMessage = "NOT FOUND";
            response.end("<h1><center>HTTP 404</center></h1><h2><center>NOT FOUND</center></h2>");
        }
    };
    Server.prototype.onGET = function (uriExpr, handler) {
        if (typeof uriExpr === "string") {
            this.pathStringMap[fixPath(uriExpr)] = handler;
        }
        else {
            this.pathRegExpMap.push({
                "expr": uriExpr,
                "handler": handler
            });
        }
        console.log("HTTP Server " + this.options.host + ":" + this.options.port + ": Registered handler 'GET " + uriExpr + "'.");
    };
    return Server;
}());
exports.Server = Server;
//# sourceMappingURL=httpServer.js.map