/// <reference path="../lrt.d.ts" />
import NodeHTTP = require("http");
import NodeURLParser = require("url");
import NodeQueryStringParser = require("querystring");
import Logger = require("../utils/logger");
import system = lrts.system;

export interface Request extends NodeHTTP.IncomingMessage {

    "query": { [key: string]: any; };

    /**
     * The path of URL.
     */
    "path": string;

    /**
     * The path of URL.
     */
    "pathMatch"?: RegExpMatchArray;
}

export type Response = NodeHTTP.ServerResponse;

export interface RequestHandler {

    (request: Request, response: NodeHTTP.ServerResponse, server?: Server): void;
}

export interface URIRegExpMap {

    "expr": RegExp;

    "handler": RequestHandler;
}

type URIExpr = string | RegExp;

/**
 * The status types for HTTP server.
 */
export enum ServerStatus {

    /**
     * Server is not working.
     */
    "IDLE" = 0,

    /**
     * Server is working normally.
     */
    "WORKING" = 1,

    /**
     * Server is waiting to be shutdown.
     */
    "CLOSING" = 2,

    /**
     * Server has been shutdown.
     */
    "CLOSED" = 3,

    /**
     * Server was in wrong situation.
     */
    "ERROR" = 0xFF
}

/**
 * Remove the last "/" of path and add first "/" if not exists.
 */
function fixPath(path: string): string {

    if (path[0] !== "/") {

        path = "/" + path;
    }

    if (path[path.length - 1] === "/") {

        return path.substr(0, path.length - 1);
    }

    return path;
}

export class Server {

    protected options: lrts.server.http.ServerOptions;

    protected serverInst: NodeHTTP.Server;

    protected serverStatus: ServerStatus;

    protected pathStringMap: { [key: string]: RequestHandler; };

    protected pathRegExpMap: URIRegExpMap[];

    constructor(options?: lrts.server.http.ServerOptions) {

        this.options = {

            "host":     "127.0.0.1",
            "port":     80,
            "backlog":  511
        };

        this.serverStatus = ServerStatus.IDLE;

        this.pathStringMap = {};

        this.pathRegExpMap = [];

        if (options) {

            for (let key in this.options) {

                options[key] && (this.options[key] = options[key]);
            }
        }

    }

    public get status(): ServerStatus {

        return this.serverStatus;
    }

    /**
     * Start the HTTP Server.
     */
    public start(cb?: ErrorCallback): void {

        let callback: Function = (function (err?: Error) {

            if (err) {

                this.serverStatus = ServerStatus.ERROR;

                console.log(`HTTP Server ${this.options.host}:${this.options.port}: Server started failed.`);

                cb && cb(err);

                return;
            }

            console.log(`HTTP Server ${this.options.host}:${this.options.port}: Server started okay.`);

            this.serverStatus = ServerStatus.WORKING;

            cb && cb(err);

        }).bind(this);

        if (this.serverStatus !== ServerStatus.IDLE) {

            return cb && cb({
                "name": "BAD-STATUS",
                "message": "Server has already started."
            });
        }

        this.serverInst = NodeHTTP.createServer(this.requestEntry.bind(this));

        this.serverInst.listen(this.options.port, this.options.host, this.options.backlog, callback);

        console.log(`HTTP Server ${this.options.host}:${this.options.port}: Server is starting...`);

    }

    public close() {

        if (this.status === ServerStatus.WORKING) {

            this.serverInst.close((function() {

                this.serverStatus = ServerStatus.CLOSED;

                console.log(`HTTP Server ${this.options.host}:${this.options.port}: Server has been shutdown.`);

            }).bind(this));

            this.serverStatus = ServerStatus.CLOSING;
        }

        console.log(`HTTP Server ${this.options.host}:${this.options.port}: Server is shutting down...`);
    }

    protected requestEntry(request: Request, response: NodeHTTP.ServerResponse): void {

        if (this.status === ServerStatus.CLOSING) {

            response.setHeader("Connection", "Close");

            response.statusCode = 503;
            response.statusMessage = "Server is Unavailable";

            response.end();

            request.socket.destroy();

            return;
        }

        let urlInfo: NodeURLParser.Url = NodeURLParser.parse(request.url);
        let handlerFound: boolean = false;

        request.path = urlInfo.pathname ? fixPath(urlInfo.pathname) : "/";

        request.query = {};

        for (let path in this.pathStringMap) {

            if (request.path === path) {

                handlerFound = true;

                setTimeout(this.pathStringMap[path], 0, request, response, this);

                break;

            }

        }

        if (!handlerFound) {

            for (let mapRule of this.pathRegExpMap) {

                if (request.pathMatch = request.path.match(mapRule.expr)) {

                    handlerFound = true;

                    setTimeout(mapRule.handler, 0, request, response, this);

                    break;

                }

            }

        }

        if (!handlerFound) {

            response.statusCode = 404;
            response.statusMessage = "NOT FOUND";

            response.end("<h1><center>HTTP 404</center></h1><h2><center>NOT FOUND</center></h2>");
        }

    }

    public onGET(uriExpr: string | RegExp, handler: RequestHandler): void {

        if (typeof uriExpr === "string") {

            this.pathStringMap[fixPath(uriExpr)] = handler;

        } else {

            this.pathRegExpMap.push({
                "expr": uriExpr,
                "handler": handler
            });

        }

        console.log(`HTTP Server ${this.options.host}:${this.options.port}: Registered handler 'GET ${uriExpr}'.`);
    }

}
